#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <cutil.h>
#include <math.h>


#define NUM_BLOCKS 4
#define NUM_THREADS 512
#define DATA_SIZE 100000000	// 100M

#define TRIES 100

#define MAX_THREADS 512
#define MAX_BLOCKS 100


// TODO: Need to do the final Kernal approach as two kernals, one to get everything into 512 blocks and a final pass to 
//     combine blocks

#include "avg_kernel.cu"

////////////////////////////////////////////////////////////////////////////////
// Helper function, returning uniformly distributed
// random float in [low, high] range
////////////////////////////////////////////////////////////////////////////////
float RandFloat(float low, float high){
    float t = (float)rand() / (float)RAND_MAX;
    return (1.0f - t) * low + t * high;
}


///////////////////////////////////////////////////////////////////////////////
// Main program
///////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    float *data;
    int i;
    float *d_data;
    float *d_sum;
    float *d_block_sum;
    int num_values;
    int data_sz;

    float cpu_time;
    float gpu_time;

    float cpu;
    float gpu;

    int numBlocks;
    int numThreads;
    int numTries;

    FILE *fp;

    cudaEvent_t start, stop;
    cudaDeviceProp deviceProp;

    numTries = TRIES;
    num_values = DATA_SIZE;
    numThreads = NUM_THREADS;
    numBlocks = NUM_BLOCKS;

    data_sz = num_values * (sizeof (float));

    fp = fopen("avg.csv", "w");

    CUT_DEVICE_INIT(argc, argv);

    cudaGetDeviceProperties(&deviceProp, 0);
    printf("\nDevice \"%s\"\n", deviceProp.name);

    // Create the timers
    CUDA_SAFE_CALL( cudaEventCreate(&start) );
    CUDA_SAFE_CALL( cudaEventCreate(&stop) );

    // Initialize the data
    CUDA_SAFE_CALL( cudaMallocHost((void **)&data, data_sz) );

    srand(123);
    for (i=0;i<num_values;i++)
        data[i] = RandFloat(-10.0, 10.0);

    //
    // First test - CPU
    //
    CUDA_SAFE_CALL( cudaEventRecord(start, 0) );

    // Figure out the sum
    cpu = 0;
    for (i=0;i<num_values;i++)
	cpu += data[i];

    CUDA_SAFE_CALL( cudaEventRecord(stop, 0) );
    CUDA_SAFE_CALL( cudaEventSynchronize(stop) );
    CUDA_SAFE_CALL( cudaEventElapsedTime(&cpu_time, start, stop) );    


    // Allocating GPU memory
    CUDA_SAFE_CALL( cudaMalloc((void **)&d_data, data_sz)   );

    // Allocating GPU memory for the block scratch
    CUDA_SAFE_CALL( cudaMalloc((void **)&d_block_sum, MAX_BLOCKS * sizeof (float))   );

    // Allocating GPU memory
    CUDA_SAFE_CALL(cudaMalloc((void **)&d_sum, MAX_THREADS * sizeof (float)));

    // Copy data to the GPU
    CUDA_SAFE_CALL( cudaMemcpy(d_data, data, data_sz, cudaMemcpyHostToDevice) );

    //
    // GPU Test
    //

	CUDA_SAFE_CALL( cudaEventRecord(start, 0) );
	CUDA_SAFE_CALL( cudaEventRecord(stop, 0) );
	CUDA_SAFE_CALL( cudaEventSynchronize(stop) );
	CUDA_SAFE_CALL( cudaEventElapsedTime(&gpu_time, start, stop) );    

    // Execute the kernal to add all the block numbers
    for (numBlocks = 1; numBlocks < (MAX_BLOCKS+1); numBlocks++)
    {
	CUDA_SAFE_CALL( cudaEventRecord(start, 0) );

	for (i=0; i< numTries; i++)
	{
	    if (numBlocks == 1)
	    {
		sumData<<<1,numThreads,4*MAX_THREADS*2>>>(d_sum, d_data, num_values);
	    }
	    else
	    {
		sumData<<<numBlocks,numThreads,4*MAX_THREADS*2>>>(d_block_sum, d_data, num_values);
		sumData<<<1,numThreads,4*MAX_THREADS*2>>>(d_sum, d_block_sum, numBlocks);
	    }

	    CUDA_SAFE_CALL( cudaMemcpy(&gpu, d_sum, sizeof (float), cudaMemcpyDeviceToHost) );
	}

	CUDA_SAFE_CALL( cudaEventRecord(stop, 0) );
	CUDA_SAFE_CALL( cudaEventSynchronize(stop) );
	CUDA_SAFE_CALL( cudaEventElapsedTime(&gpu_time, start, stop) );    

	fprintf(fp,"%d,%f,%f,%f\n", numBlocks, gpu_time / (float) numTries, cpu_time / gpu_time, gpu);
    }

    printf("CPU Sum:\t%g\n", cpu);
    printf("GPU Sum:\t%g\n", gpu);
    printf("\n");

    fclose(fp);

    CUDA_SAFE_CALL(  cudaFreeHost(data) );

    CUDA_SAFE_CALL( cudaEventDestroy(start) );
    CUDA_SAFE_CALL( cudaEventDestroy(stop) );

    CUT_EXIT(1, argv);
}

